@@ -7,6 +7,10 @@
 
 #include "links.h"
 
+#ifdef HAIKU
+#include <FindDirectory.h>
+#endif
+
 unsigned char system_name[MAX_STR_LEN];
 
 static void get_system_name(void)
@@ -21,20 +25,20 @@ static void get_system_name(void)
 		int rs;
 		memset(&name, 0, sizeof name);
 		EINTRLOOP(rs, uname(&name));
-		if (!rs) {
+		if (rs >= 0) {
 #ifdef OPENVMS
 			unsigned char * volatile p;
 #endif
 			unsigned char *str = init_str();
 			int l = 0;
 			add_to_str(&str, &l, cast_uchar name.sysname);
-			add_to_str(&str, &l, cast_uchar " ");
+			add_chr_to_str(&str, &l, ' ');
 #ifdef OPENVMS
 			add_to_str(&str, &l, cast_uchar name.version);
 #else
 			add_to_str(&str, &l, cast_uchar name.release);
 #endif
-			add_to_str(&str, &l, cast_uchar " ");
+			add_chr_to_str(&str, &l, ' ');
 #ifdef OPENVMS
 			p = cast_uchar name.nodename + sizeof(name.nodename);
 			if ((unsigned char *)(&name + 1) - p >= 16 && memchr(cast_const_char p, 0, 16))
@@ -396,8 +400,7 @@ static void get_compiler_name(void)
 static void do_exit(int x)
 {
 	/* avoid compiler warnings about unreachable code */
-	static volatile int zero = 0;
-	if (!zero) exit(x);
+	if (value_1()) exit(x);
 }
 
 struct option {
@@ -450,7 +453,7 @@ static unsigned char *p_arse_options(int argc, unsigned char *argv[], struct opt
 	return cast_uchar "";
 }
 
-static unsigned char *get_token(unsigned char **line)
+unsigned char *get_token(unsigned char **line)
 {
 	unsigned char *s = NULL;
 	int l = 0;
@@ -478,19 +481,20 @@ static unsigned char *get_token(unsigned char **line)
 	return s;
 }
 
-static int get_token_num(unsigned char **line)
+static int get_token_num(unsigned char **line, int min, int max, int *result)
 {
 	long l;
 	unsigned char *end;
 	unsigned char *t = get_token(line);
 	if (!t) return -1;
 	l = strtolx(t, &end);
-	if (*end || end == t || l < 0 || l != (int)l) {
+	if (*end || end == t || l < min || l > max) {
 		mem_free(t);
-		return -1;
+		return -2;
 	}
 	mem_free(t);
-	return (int)l;
+	*result = (int)l;
+	return 0;
 }
 
 static void parse_config_file(unsigned char *name, unsigned char *file, struct option **opt)
@@ -528,7 +532,7 @@ static void parse_config_file(unsigned char *name, unsigned char *file, struct o
 		for (op = opt; (options = *op); op++)
 			for (i = 0; options[i].p; i++) if (options[i].cfg_name && (size_t)nl == strlen(cast_const_char options[i].cfg_name) && !casecmp(tok, cast_uchar options[i].cfg_name, nl)) {
 				unsigned char *o = memacpy(p, pl);
-				if ((e = options[i].rd_cfg(&options[i], o))) {
+				if (options[i].rd_cfg && (e = options[i].rd_cfg(&options[i], o))) {
 					if (e[0]) fprintf(stderr, "Error parsing config file %s, line %d: %s\n", name, line, e), err = 1;
 				}
 				mem_free(o);
@@ -559,29 +563,41 @@ static unsigned char *create_config_string(struct option *options)
 	return s;
 }
 
-#define FILE_BUF	1024
-
-static unsigned char cfg_buffer[FILE_BUF];
-
 unsigned char *read_config_file(unsigned char *name)
 {
 	int h, r;
 	int l = 0;
-	unsigned char *s;
+	unsigned char *cfg_buffer, *s;
 	int rs;
 	h = c_open(name, O_RDONLY | O_NOCTTY);
 	if (h == -1) return NULL;
 	s = init_str();
-	while ((r = hard_read(h, cfg_buffer, FILE_BUF)) > 0) {
+	cfg_buffer = mem_alloc(page_size);
+	while ((r = hard_read(h, cfg_buffer, page_size)) > 0) {
 		int i;
 		for (i = 0; i < r; i++) if (!cfg_buffer[i]) cfg_buffer[i] = ' ';
 		add_bytes_to_str(&s, &l, cfg_buffer, r);
 	}
+	mem_free(cfg_buffer);
 	if (r == -1) mem_free(s), s = NULL;
 	EINTRLOOP(rs, close(h));
 	return s;
 }
 
+void delete_config_file(unsigned char *name)
+{
+	int rs;
+#if defined(OPENVMS)
+	/* delete all versions of the file */
+	int count = 0;
+	do {
+		EINTRLOOP(rs, unlink(cast_const_char name));
+	} while (!rs && ++count < 65536);
+#else
+	EINTRLOOP(rs, unlink(cast_const_char name));
+#endif
+}
+
 int write_to_config_file(unsigned char *name, unsigned char *c, int do_sync)
 {
 	int rr;
@@ -627,20 +643,17 @@ try_new_count:
 			err = errno;
 			goto close_unlink_err;
 		}
+#ifdef DOS
+		_flush_disk_cache();
+#endif
 	}
 	EINTRLOOP(rs, close(h));
 	if (rs) {
 		err = errno;
 		goto unlink_err;
 	}
-#if defined(OPENVMS)
-	/* delete all versions of the file */
-	count = 0;
-	do {
-		EINTRLOOP(rs, unlink(cast_const_char name));
-	} while (!rs && ++count < 65536);
-#elif !defined(RENAME_OVER_EXISTING_FILES)
-	EINTRLOOP(rs, unlink(cast_const_char name));
+#if defined(OPENVMS) || !defined(RENAME_OVER_EXISTING_FILES)
+	delete_config_file(name);
 #endif
 	EINTRLOOP(rs, rename(cast_const_char tmp_name, cast_const_char name));
 	if (rs) {
@@ -648,8 +661,10 @@ try_new_count:
 		goto unlink_err;
 	}
 	mem_free(tmp_name);
-#if !defined(OS2) && !defined(OPENVMS)
 	if (do_sync) {
+#if defined(DOS)
+		_flush_disk_cache();
+#elif !defined(OS2) && !defined(OPENVMS)
 		unsigned char *e, *le;
 		tmp_name = stracpy(name);
 		le = tmp_name;
@@ -672,8 +687,8 @@ try_new_count:
 			EINTRLOOP(rs, close(h));
 		}
 		mem_free(tmp_name);
-	}
 #endif
+	}
 
 	return 0;
 
@@ -746,6 +761,13 @@ skip_path_conv:;
 			}
 		}
 	}
+#endif
+#ifdef HAIKU
+	home = mem_alloc(B_FILE_NAME_LENGTH);
+	if (find_directory(B_USER_SETTINGS_DIRECTORY, -1, 1, cast_char home, B_FILE_NAME_LENGTH) != B_OK) {
+		mem_free(home);
+		home = NULL;
+	}
 #endif
 	if (!home) home = stracpy(cast_uchar getenv("HOME"));
 #ifdef WIN
@@ -793,7 +815,9 @@ skip_path_conv:;
 		}
 	} else {
 		add_dot_links:
-#ifdef OPENVMS
+#if defined(DOS)
+		add_to_strn(&home_links, cast_uchar "links.cfg");
+#elif defined(OPENVMS) || defined(HAIKU)
 		add_to_strn(&home_links, cast_uchar "links");
 #else
 		add_to_strn(&home_links, cast_uchar ".links");
@@ -889,7 +913,7 @@ static int write_config_data(unsigned char *prefix, unsigned char *name, struct
 	}
 	add_to_strn(&config_file, name);
 	if ((err = write_to_config_file(config_file, c, 1))) {
-		if (term) msg_box(term, NULL, TEXT_(T_CONFIG_ERROR), AL_CENTER, TEXT_(T_UNABLE_TO_WRITE_TO_CONFIG_FILE), cast_uchar ": ", get_err_msg(err), MSG_BOX_END, NULL, 1, TEXT_(T_CANCEL), msg_box_null, B_ENTER | B_ESC);
+		if (term) msg_box(term, NULL, TEXT_(T_CONFIG_ERROR), AL_CENTER, TEXT_(T_UNABLE_TO_WRITE_TO_CONFIG_FILE), cast_uchar ": ", get_err_msg(err, term), MSG_BOX_END, NULL, 1, TEXT_(T_CANCEL), msg_box_null, B_ENTER | B_ESC);
 		mem_free(c);
 		mem_free(config_file);
 		return -1;
@@ -903,10 +927,10 @@ static void add_nm(struct option *o, unsigned char **s, int *l)
 {
 	if (*l) add_to_str(s, l, cast_uchar NEWLINE);
 	add_to_str(s, l, cast_uchar o->cfg_name);
-	add_to_str(s, l, cast_uchar " ");
+	add_chr_to_str(s, l, ' ');
 }
 
-static void add_quoted_to_str(unsigned char **s, int *l, unsigned char *q)
+void add_quoted_to_str(unsigned char **s, int *l, unsigned char *q)
 {
 	add_chr_to_str(s, l, '"');
 	while (*q) {
@@ -946,7 +970,7 @@ static void num_wr(struct option *o, unsigned char **s, int *l)
 static unsigned char *dbl_rd(struct option *o, unsigned char *c)
 {
 	unsigned char *tok = get_token(&c);
-	unsigned char *end;
+	char *end;
 	double d;
 
 	if (!tok) return cast_uchar "Missing argument";
@@ -956,13 +980,13 @@ static unsigned char *dbl_rd(struct option *o, unsigned char *c)
 		return cast_uchar "Number is too long";
 	}
 
-	d = strtod(cast_const_char tok, (char **)(void *)&end);
+	d = strtod(cast_const_char tok, &end);
 
-	if (*end) {
+	if (*end || tok[strspn(cast_const_char tok, "0123456789.eE+-")]) {
 		mem_free(tok);
 		return cast_uchar "Number expected";
 	}
-	if (d < 0 || d > o->max || 100*d < o->min || 100*d > o->max) {
+	if (d < 0 || d > o->max || 100 * d < o->min || 100 * d > o->max) {
 		mem_free(tok);
 		return cast_uchar "Out of range";
 	}
@@ -984,7 +1008,7 @@ static unsigned char *str_rd(struct option *o, unsigned char *c)
 {
 	unsigned char *tok = get_token(&c);
 	unsigned char *e = NULL;
-	if (!tok) return NULL;
+	if (!tok) tok = stracpy(cast_uchar "");
 	if (strlen(cast_const_char tok) + 1 > (size_t)o->max) e = cast_uchar "String too long";
 	else strcpy(cast_char o->ptr, cast_const_char tok);
 	mem_free(tok);
@@ -1049,8 +1073,8 @@ static void lang_wr(struct option *o, unsigned char **s, int *l)
 
 static int getnum(unsigned char *s, int *n, int r1, int r2)
 {
-	unsigned char *e;
-	long l = strtol(cast_const_char s, (char **)(void *)&e, 10);
+	char *e;
+	long l = strtol(cast_const_char s, &e, 10);
 	if (*e || !*s) return -1;
 	if (l < r1 || l >= r2) return -1;
 	*n = (int)l;
@@ -1126,13 +1150,13 @@ static void type_wr(struct option *o, unsigned char **s, int *l)
 		struct assoc *as = get_struct(a, struct assoc, head);
 		add_nm(o, s, l);
 		add_quoted_to_str(s, l, as->label);
-		add_to_str(s, l, cast_uchar " ");
+		add_chr_to_str(s, l, ' ');
 		add_quoted_to_str(s, l, as->ct);
-		add_to_str(s, l, cast_uchar " ");
+		add_chr_to_str(s, l, ' ');
 		add_quoted_to_str(s, l, as->prog);
-		add_to_str(s, l, cast_uchar " ");
+		add_chr_to_str(s, l, ' ');
 		add_num_to_str(s, l, (!!as->cons) + (!!as->xwin) * 2 + (!!as->ask) * 4 + (!as->block) * 8 + (!!as->block) * 16 + (!!as->accept_http) * 32 + (!!as->accept_ftp) * 64);
-		add_to_str(s, l, cast_uchar " ");
+		add_chr_to_str(s, l, ' ');
 		add_num_to_str(s, l, as->system);
 	}
 }
@@ -1160,7 +1184,7 @@ static void ext_wr(struct option *o, unsigned char **s, int *l)
 		struct extension *e = get_struct(a, struct extension, head);
 		add_nm(o, s, l);
 		add_quoted_to_str(s, l, e->ext);
-		add_to_str(s, l, cast_uchar " ");
+		add_chr_to_str(s, l, ' ');
 		add_quoted_to_str(s, l, e->ct);
 	}
 }
@@ -1191,7 +1215,7 @@ static void prog_wr(struct option *o, unsigned char **s, int *l)
 		if (!*a->prog) continue;
 		add_nm(o, s, l);
 		add_quoted_to_str(s, l, a->prog);
-		add_to_str(s, l, cast_uchar " ");
+		add_chr_to_str(s, l, ' ');
 		add_num_to_str(s, l, a->system);
 	}
 }
@@ -1200,12 +1224,12 @@ static unsigned char *term_rd(struct option *o, unsigned char *c)
 {
 	struct term_spec *ts;
 	unsigned char *w;
-	int i, l;
+	int i, r;
 	if (!(w = get_token(&c))) goto err;
 	ts = new_term_spec(w);
 	mem_free(w);
 	if (!(w = get_token(&c))) goto err;
-	if (strlen(cast_const_char w) != 1 || w[0] < '0' || w[0] > '4') goto err_f;
+	if (strlen(cast_const_char w) != 1 || w[0] < '0' || w[0] > '5') goto err_f;
 	ts->mode = w[0] - '0';
 	mem_free(w);
 	if (!(w = get_token(&c))) goto err;
@@ -1232,27 +1256,20 @@ static unsigned char *term_rd(struct option *o, unsigned char *c)
 #endif
 	ts->character_set = i;
 	mem_free(w);
-	l = get_token_num(&c);
-	if (l < 0) goto ret;
-	if (l > 999) goto err;
-	ts->left_margin = l;
-	l = get_token_num(&c);
-	if (l < 0) goto err;
-	if (l > 999) goto err;
-	ts->right_margin = l;
-	l = get_token_num(&c);
-	if (l < 0) goto err;
-	if (l > 999) goto err;
-	ts->top_margin = l;
-	l = get_token_num(&c);
-	if (l < 0) goto err;
-	if (l > 999) goto err;
-	ts->bottom_margin = l;
-	ret:
+	r = get_token_num(&c, 0, 999, &ts->left_margin);
+	if (r == -1) goto ret;
+	if (r < 0) goto err;
+	r = get_token_num(&c, 0, 999, &ts->right_margin);
+	if (r < 0) goto err;
+	r = get_token_num(&c, 0, 999, &ts->top_margin);
+	if (r < 0) goto err;
+	r = get_token_num(&c, 0, 999, &ts->bottom_margin);
+	if (r < 0) goto err;
+ret:
 	return NULL;
-	err_f:
+err_f:
 	mem_free(w);
-	err:
+err:
 	return cast_uchar "Error reading terminal specification";
 }
 
@@ -1307,23 +1324,23 @@ static void term_wr(struct option *o, unsigned char **s, int *l)
 	foreachback(struct term_spec, ts, lts, term_specs) {
 		add_nm(o, s, l);
 		add_quoted_to_str(s, l, ts->term);
-		add_to_str(s, l, cast_uchar " ");
+		add_chr_to_str(s, l, ' ');
 		add_num_to_str(s, l, ts->mode);
-		add_to_str(s, l, cast_uchar " ");
+		add_chr_to_str(s, l, ' ');
 		add_num_to_str(s, l, !!ts->m11_hack + !!ts->braille * 2);
-		add_to_str(s, l, cast_uchar " ");
+		add_chr_to_str(s, l, ' ');
 		add_num_to_str(s, l, !!ts->col + !!ts->restrict_852 * 2 + !!ts->block_cursor * 4);
-		add_to_str(s, l, cast_uchar " ");
+		add_chr_to_str(s, l, ' ');
 		if (ts->character_set == -1) add_to_str(s, l, cast_uchar "default");
 		else add_to_str(s, l, get_cp_mime_name(ts->character_set));
 		if (ts->left_margin || ts->right_margin || ts->top_margin || ts->bottom_margin) {
-			add_to_str(s, l, cast_uchar " ");
+			add_chr_to_str(s, l, ' ');
 			add_num_to_str(s, l, ts->left_margin);
-			add_to_str(s, l, cast_uchar " ");
+			add_chr_to_str(s, l, ' ');
 			add_num_to_str(s, l, ts->right_margin);
-			add_to_str(s, l, cast_uchar " ");
+			add_chr_to_str(s, l, ' ');
 			add_num_to_str(s, l, ts->top_margin);
-			add_to_str(s, l, cast_uchar " ");
+			add_chr_to_str(s, l, ' ');
 			add_num_to_str(s, l, ts->bottom_margin);
 		}
 	}
@@ -1340,9 +1357,9 @@ struct driver_param *get_driver_param(unsigned char *n)
 	sl = strlen(cast_const_char n);
 	if (sl > MAXINT - sizeof(struct driver_param)) overalloc();
 	dp = mem_calloc(sizeof(struct driver_param) + sl);
-	dp->kbd_codepage = -1;
 	strcpy(cast_char dp->name, cast_const_char n);
-	dp->shell = mem_calloc(1);
+	dp->kbd_codepage = -1;
+	dp->palette_mode = 0;
 	dp->nosave = 1;
 	add_to_list(driver_params, dp);
 	return dp;
@@ -1351,44 +1368,39 @@ struct driver_param *get_driver_param(unsigned char *n)
 static unsigned char *dp_rd(struct option *o, unsigned char *c)
 {
 	int cc;
-	unsigned char *n, *param, *cp, *shell;
+	unsigned char *n, *param, *cp;
 	struct driver_param *dp;
-	if (!(n = get_token(&c))) goto err;
+	if (!(n = get_token(&c))) {
+		goto err;
+	}
+	dp = get_driver_param(n);
+	mem_free(n);
 	if (!(param = get_token(&c))) {
-		mem_free(n);
 		goto err;
 	}
-	if (!(shell = get_token(&c))){
-		mem_free(n);
-		mem_free(param);
+	if (dp->param) mem_free(dp->param);
+	dp->param = param;
+	if (!(param = get_token(&c))){
 		goto err;
 	}
+	safe_strncpy(dp->shell_term, param, MAX_STR_LEN);
+	mem_free(param);
 	if (!(cp = get_token(&c))) {
-		mem_free(n);
-		mem_free(param);
-		mem_free(shell);
 		goto err;
 	}
 	if (!casestrcmp(cp, cast_uchar "default")) {
 		cc = -1;
 	} else if ((cc = get_cp_index(cp)) == -1) {
-		mem_free(n);
-		mem_free(param);
-		mem_free(shell);
 		mem_free(cp);
 		goto err;
 	}
-	dp = get_driver_param(n);
-	dp->kbd_codepage = cc;
-	if (dp->param) mem_free(dp->param);
-	dp->param = param;
-	if (dp->shell) mem_free(dp->shell);
-	dp->shell = shell;
-	dp->nosave = 0;
 	mem_free(cp);
-	mem_free(n);
+	dp->kbd_codepage = cc;
+	cc = get_token_num(&c, 0, 1, &dp->palette_mode);
+	if (cc == -2) goto err;
+	dp->nosave = 0;
 	return NULL;
-	err:
+err:
 	return cast_uchar "Error reading driver mode specification";
 }
 
@@ -1397,40 +1409,23 @@ static void dp_wr(struct option *o, unsigned char **s, int *l)
 	struct driver_param *dp;
 	struct list_head *ldp;
 	foreachback(struct driver_param, dp, ldp, driver_params) {
-		if ((!dp->param || !*dp->param) && dp->kbd_codepage == -1 && !*dp->shell) continue;
+		if ((!dp->param || !*dp->param) && !*dp->shell_term && dp->kbd_codepage == -1 && !dp->palette_mode) continue;
 		if (dp->nosave) continue;
 		add_nm(o, s, l);
 		add_quoted_to_str(s, l, dp->name);
-		add_to_str(s, l, cast_uchar " ");
+		add_chr_to_str(s, l, ' ');
 		add_quoted_to_str(s, l, dp->param ? dp->param : (unsigned char*)"");
-		add_to_str(s, l, cast_uchar " ");
-		add_quoted_to_str(s, l, dp->shell);
-		add_to_str(s, l, cast_uchar " ");
+		add_chr_to_str(s, l, ' ');
+		add_quoted_to_str(s, l, dp->shell_term);
+		add_chr_to_str(s, l, ' ');
 		if (dp->kbd_codepage == -1) add_to_str(s, l, cast_uchar "default");
 		else add_to_str(s, l, get_cp_mime_name(dp->kbd_codepage));
+		add_chr_to_str(s, l, ' ');
+		add_num_to_str(s, l, dp->palette_mode);
+		/* pokud se sem neco prida, opravit podminku na zacatku cyklu */
 	}
 }
 
-static unsigned char *ip_rd(struct option *o, unsigned char *c)
-{
-	unsigned char *e;
-	e = str_rd(o, c);
-	if (e) return e;
-	if (*(unsigned char *)o->ptr && numeric_ip_address(o->ptr, NULL) == -1) return cast_uchar "Invalid IP address";
-	return NULL;
-}
-
-static unsigned char *ipv6_rd(struct option *o, unsigned char *c)
-{
-	unsigned char *e;
-	e = str_rd(o, c);
-	if (e) return e;
-#ifdef SUPPORT_IPV6
-	if (*(unsigned char *)o->ptr && numeric_ipv6_address(o->ptr, NULL, NULL) == -1) return cast_uchar "Invalid IPv6 address";
-#endif
-	return NULL;
-}
-
 static unsigned char *gen_cmd(struct option *o, unsigned char ***argv, int *argc)
 {
 	unsigned char *e;
@@ -1440,7 +1435,7 @@ static unsigned char *gen_cmd(struct option *o, unsigned char ***argv, int *argc
 	e = init_str();
 	l = 0;
 	add_quoted_to_str(&e, &l, **argv);
-	r = o->rd_cfg(o, e);
+	r = o->rd_cfg ? o->rd_cfg(o, e) : 0;
 	mem_free(e);
 	if (r) return r;
 	(*argv)++; (*argc)--;
@@ -1647,13 +1642,13 @@ fprintf(stdout, "%s%s%s%s%s%s\n",
 "    (default: 600)\n"
 "\n"
 " -timeout-when-trying-multiple-addresses <sec>\n"
-"  Timeout for connection when trying multiple addresses.\n"
+"  Timeout for connection when trying multiple addresses or keepalive connection.\n"
 "    (default: 3)\n"
 "\n"
-" -bind-address <ip address>\n"
+" -bind-address <ip address or interface name>\n"
 "  Use a specific local IP address.\n"
 "\n"
-" -bind-address-ipv6 <ipv6 address>\n"
+" -bind-address-ipv6 <ipv6 address or interface name>\n"
 "  Use a specific local IPv6 address.\n"
 "\n"
 " -no-libevent\n"
@@ -1662,9 +1657,6 @@ fprintf(stdout, "%s%s%s%s%s%s\n",
 " -no-openmp\n"
 "  Don't use OpenMP.\n"
 "\n"
-" -async-dns <0>/<1>\n"
-"  Asynchronous DNS resolver on(1)/off(0).\n"
-"\n"
 " -download-utime <0>/<1>\n"
 "  Set time of downloaded files to last modification time reported by server.\n"
 "\n"
@@ -1752,6 +1744,13 @@ fprintf(stdout, "%s%s%s%s%s%s\n",
 "\n"
 " -ssl.client-cert-password <text>\n"
 "  Password for the user private key.\n"
+"\n"
+" -async-dns <0>/<1>\n"
+"  Asynchronous DNS resolver on(1)/off(0).\n"
+"\n"
+" -dns-over-https <url>\n"
+"  DNS over HTTPS url.\n"
+"    (default: blank)\n"
 "\n"),(
 " -http-bugs.http10 <0>/<1>\n"
 "    (default 0)\n"
@@ -1764,15 +1763,6 @@ fprintf(stdout, "%s%s%s%s%s%s\n",
 "  \"1\" defaults to using list of servers that have broken HTTP/1.1 support.\n"
 "     When links finds such server, it will retry the request with HTTP/1.0.\n"
 "\n"
-" -http-bugs.bug-302-redirect <0>/<1>\n"
-"    (default 1)\n"
-"  Process 302 redirect in a way that is incompatible with RFC1945 and RFC2068,\n"
-"    but the same as Netscape and MSIE. Many pages depend on it.\n"
-"\n"
-" -http-bugs.bug-post-no-keepalive <0>/<1>\n"
-"    (default 0)\n"
-"  No keepalive connection after post requests. For some buggy servers.\n"
-"\n"
 " -http-bugs.bug-no-accept-charset <0>/<1>\n"
 "    (default 0)\n"
 "  Do not send Accept-Charset field of HTTP header. Because it is too long\n"
@@ -1823,11 +1813,6 @@ fprintf(stdout, "%s%s%s%s%s%s\n",
 " -ftp.use-eprt-epsv <0>/<1>\n"
 "  Use EPRT and EPSV commands instead of PORT and PASV.\n"
 "\n"
-" -ftp.fast <0>/<1>\n"
-"  Send more ftp commands simultaneously. Faster response when\n"
-"    browsing ftp directories, but it is incompatible with RFC\n"
-"    and some servers don't like it.\n"
-"\n"
 " -ftp.set-iptos <0>/<1>\n"
 "  Set IP Type-of-service to high throughput on ftp connections.\n"
 "\n"
@@ -1863,6 +1848,15 @@ fprintf(stdout, "%s%s%s%s%s%s\n",
 " -save-url-history <0>/<1>\n"
 "  Save URL history on exit.\n"
 "\n"
+" -enable-cookies <0>/<1>\n"
+"  Enable cookiest.\n"
+"\n"
+" -save-cookies <0>/<1>\n"
+"  Save cookies on exit.\n"
+"\n"
+" -max-cookie-age <fp-value>\n"
+"  Maximum cookie age in days. 0 - unlimited\n"
+"\n"
 " -display-red-gamma <fp-value>\n"
 "  Red gamma of display.\n"
 "    (default 2.2)\n"
@@ -1903,6 +1897,34 @@ fprintf(stdout, "%s%s%s%s%s%s\n",
 "    (valid for svgalib and framebuffer).\n"
 "    Overwriting may or may not be faster, depending on hardware.\n"
 "\n"
+#ifdef HAVE_FREETYPE
+" -font <file>\n"
+"  Font file for normal text\n"
+"\n"
+" -font-bold <file>\n"
+"  Font file for bold text\n"
+"\n"
+" -font-monospaced <file>\n"
+"  Font file for monospaced text\n"
+"\n"
+" -font-monospaced-bold <file>\n"
+"  Font file for monospaced bold text\n"
+"\n"
+#ifdef USE_ITALIC
+" -font-italic <file>\n"
+"  Font file for normal text\n"
+"\n"
+" -font-italic-bold <file>\n"
+"  Font file for bold text\n"
+"\n"
+" -font-monospaced-italic <file>\n"
+"  Font file for monospaced text\n"
+"\n"
+" -font-monospaced-italic-bold <file>\n"
+"  Font file for monospaced bold text\n"
+"\n"
+#endif
+#endif
 #ifdef JS
 " -enable-javascript <0>/<1>\n"
 "  Enable javascript.\n"
@@ -1980,16 +2002,16 @@ fprintf(stdout, "%s%s%s%s%s%s\n",
 " -html-user-font-size <size>\n"
 "  Size of font on pages in graphics mode.\n"
 "\n"
-" -html-t-text-color <0>-<15>\n"
+" -html-text-color <0>-<15>\n"
 "  Text color in text mode.\n"
 "\n"
-" -html-t-link-color <0>-<15>\n"
+" -html-link-color <0>-<15>\n"
 "  Link color in text mode.\n"
 "\n"
-" -html-t-background-color <0>-<7>\n"
+" -html-background-color <0>-<7>\n"
 "  Background color in text mode.\n"
 "\n"
-" -html-t-ignore-document-color <0>/<1>\n"
+" -html-ignore-document-color <0>/<1>\n"
 "  Ignore colors specified in html document in text mode.\n"
 "\n"
 " -html-g-text-color 0xRRGGBB\n"
@@ -2060,7 +2082,6 @@ void end_config(void)
 	struct list_head *ldp;
 	foreach(struct driver_param, dp, ldp, driver_params) {
 		if (dp->param) mem_free(dp->param);
-		if (dp->shell) mem_free(dp->shell);
 	}
 	free_list(struct driver_param, driver_params);
 	if (links_home) mem_free(links_home);
@@ -2102,7 +2123,6 @@ int unrestartable_receive_timeout = 600;
 int timeout_multiple_addresses = 3;
 unsigned char bind_ip_address[16] = "";
 unsigned char bind_ipv6_address[INET6_ADDRSTRLEN] = "";
-int async_lookup = 1;
 int download_utime = 0;
 
 int max_format_cache_entries = 5;
@@ -2120,33 +2140,46 @@ struct ssl_options ssl_options = { SSL_WARN_ON_INVALID_CERTIFICATE,
 	0,
 #endif
 	"", "", "" };
-struct http_options http_options = { 0, 1, 1, 0, 0, 0, 0, { 0, 0, REFERER_REAL_SAME_SERVER, "", "", "" } };
-struct ftp_options ftp_options = { "somebody@host.domain", 0, 0, 0, 1 };
+int async_lookup = 1;
+unsigned char dns_over_https[MAX_STR_LEN] = "";
+struct http_options http_options = { 0, 1, 0, 0, 0, { 0, 0, REFERER_REAL_SAME_SERVER, "", "", "" } };
+struct ftp_options ftp_options = { "somebody@host.domain", 1, 0, 1 };
 struct smb_options smb_options = { 0 };
 
 unsigned char download_dir[MAX_STR_LEN] = "";
 
-static int sink;
-int js_enable=1;   /* 0=disable javascript */
-int js_verbose_errors=0;   /* 1=create dialog on every javascript error, 0=be quiet and continue */
-int js_verbose_warnings=0;   /* 1=create dialog on every javascript warning, 0=be quiet and continue */
-int js_all_conversions=1;
-int js_global_resolve=1;	/* resolvovani v globalnim adresnim prostoru, kdyz BFU vomitne document */
-int js_manual_confirmation=1; /* !0==annoying dialog on every goto url etc. */
-int js_fun_depth=100;
-int js_memory_limit=5*1024;  /* in kilobytes, should be in range 1M-20M (1MB=1024*1024B) */
+int js_enable = 1;   /* 0=disable javascript */
+int js_verbose_errors = 0;   /* 1=create dialog on every javascript error, 0=be quiet and continue */
+int js_verbose_warnings = 0;   /* 1=create dialog on every javascript warning, 0=be quiet and continue */
+int js_all_conversions = 1;
+int js_global_resolve = 1;	/* resolvovani v globalnim adresnim prostoru, kdyz BFU vomitne document */
+int js_manual_confirmation = 1; /* !0==annoying dialog on every goto url etc. */
+int js_fun_depth = 100;
+int js_memory_limit = 5 * 1024;  /* in kilobytes, should be in range 1M-20M (1MB=1024*1024B) */
 
-double display_red_gamma=2.2; /* Red gamma exponent of the display */
-double display_green_gamma=2.2; /* Green gamma exponent of the display */
-double display_blue_gamma=2.2; /* Blue gamma exponent of the display */
-double user_gamma=1.0; /* 1.0 for 64 lx. This is the number user directly changes in the menu */
-double bfu_aspect=1; /* 0.1 to 10.0, 1.0 default. >1 makes circle wider */
-int display_optimize=0;	/*0=CRT, 1=LCD RGB, 2=LCD BGR */
-int dither_letters=1;
-int dither_images=1;
-int gamma_bits=2;	/*0 --- 8, 1 --- 16, 2 --- auto */
+double display_red_gamma = 2.2; /* Red gamma exponent of the display */
+double display_green_gamma = 2.2; /* Green gamma exponent of the display */
+double display_blue_gamma = 2.2; /* Blue gamma exponent of the display */
+double user_gamma = 1.0; /* 1.0 for 64 lx. This is the number user directly changes in the menu */
+double bfu_aspect = 1; /* 0.1 to 10.0, 1.0 default. >1 makes circle wider */
+int display_optimize = 0;	/*0=CRT, 1=LCD RGB, 2=LCD BGR */
+int dither_letters = 1;
+int dither_images = 1;
+int gamma_bits = 2;	/*0 --- 8, 1 --- 16, 2 --- auto */
 int overwrite_instead_of_scroll = 1;
 
+unsigned char font_file[MAX_STR_LEN] = "";
+unsigned char font_file_b[MAX_STR_LEN] = "";
+unsigned char font_file_m[MAX_STR_LEN] = "";
+unsigned char font_file_m_b[MAX_STR_LEN] = "";
+
+#ifdef USE_ITALIC
+unsigned char font_file_i[MAX_STR_LEN] = "";
+unsigned char font_file_i_b[MAX_STR_LEN] = "";
+unsigned char font_file_i_m[MAX_STR_LEN] = "";
+unsigned char font_file_i_m_b[MAX_STR_LEN] = "";
+#endif
+
 
 int menu_font_size = G_BFU_DEFAULT_FONT_SIZE;
 
@@ -2156,11 +2189,15 @@ unsigned G_SCROLL_BAR_AREA_COLOR = G_DEFAULT_SCROLL_BAR_AREA_COLOR;
 unsigned G_SCROLL_BAR_BAR_COLOR = G_DEFAULT_SCROLL_BAR_BAR_COLOR;
 unsigned G_SCROLL_BAR_FRAME_COLOR = G_DEFAULT_SCROLL_BAR_FRAME_COLOR;
 
-unsigned char bookmarks_file[MAX_STR_LEN]="";
-int bookmarks_codepage=0; /* changed to utf-8 table in init_charset() */
+unsigned char bookmarks_file[MAX_STR_LEN] = "";
+int bookmarks_codepage = 0; /* changed to utf-8 table in init_charset() */
 
 int save_history = 1;
 
+int enable_cookies = 1;
+int save_cookies = 1;
+double max_cookie_age = 0;
+
 struct document_setup dds = {
 	0, /* assumed codepage */
 	0, /* ignore codepage from server */
@@ -2179,7 +2216,7 @@ struct document_setup dds = {
 	0, /* porn enable */
 	0, /* target in new window */
 	7, /* t text color */
-	15, /* t link color */
+	3, /* t link color */
 	0, /* t background color */
 	0, /* t ignore document color */
 	0x000000, /* g text color */
@@ -2217,11 +2254,10 @@ static struct option links_options[] = {
 	{1, gen_cmd, num_rd, num_wr, 1, 9999, &receive_timeout, "receive_timeout", "receive-timeout"},
 	{1, gen_cmd, num_rd, num_wr, 1, 9999, &unrestartable_receive_timeout, "unrestartable_receive_timeout", "unrestartable-receive-timeout"},
 	{1, gen_cmd, num_rd, num_wr, 1, 999, &timeout_multiple_addresses, "timeout_when_trying_multiple_addresses", "timeout-when-trying-multiple-addresses"},
-	{1, gen_cmd, ip_rd, str_wr, 0, 16, bind_ip_address, "bind_address", "bind-address"},
-	{1, gen_cmd, ipv6_rd, str_wr, 0, INET6_ADDRSTRLEN, bind_ipv6_address, "bind_address_ipv6", "bind-address-ipv6"},
+	{1, gen_cmd, str_rd, str_wr, 0, 16, bind_ip_address, "bind_address", "bind-address"},
+	{1, gen_cmd, str_rd, str_wr, 0, INET6_ADDRSTRLEN, bind_ipv6_address, "bind_address_ipv6", "bind-address-ipv6"},
 	{1, set_cmd, NULL, NULL, 0, 0, &disable_libevent, NULL, "no-libevent"},
 	{1, set_cmd, NULL, NULL, 0, 0, &disable_openmp, NULL, "no-openmp"},
-	{1, gen_cmd, num_rd, num_wr, 0, 1, &async_lookup, "async_dns", "async-dns"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &download_utime, "download_utime", "download-utime"},
 	{1, gen_cmd, num_rd, num_wr, 0, 999, &max_format_cache_entries, "format_cache_size", "format-cache-size"},
 	{1, gen_cmd, num_rd, num_wr, 0, MAXINT, &memory_cache_size, "memory_cache_size", "memory-cache-size"},
@@ -2244,10 +2280,12 @@ static struct option links_options[] = {
 	{1, gen_cmd, str_rd, NULL, 0, MAX_STR_LEN, &ssl_options.client_cert_password, NULL, "ssl.client-cert-password"},
 	{1, gen_cmd, str_rd, NULL, 0, MAX_STR_LEN, &ssl_options.client_cert_key, "client_cert_key", "http.client_cert_key"}, /* backward compatibility with Debian patches */
 	{1, gen_cmd, str_rd, NULL, 0, MAX_STR_LEN, &ssl_options.client_cert_crt, "client_cert_crt", "http.client_cert_crt"}, /* backward compatibility with Debian patches */
+	{1, gen_cmd, num_rd, num_wr, 0, 1, &async_lookup, "async_dns", "async-dns"},
+	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, dns_over_https, "dns_over_https", "dns-over-https"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &http_options.http10, "http_bugs.http10", "http-bugs.http10"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &http_options.allow_blacklist, "http_bugs.allow_blacklist", "http-bugs.allow-blacklist"},
-	{1, gen_cmd, num_rd, num_wr, 0, 1, &http_options.bug_302_redirect, "http_bugs.bug_302_redirect", "http-bugs.bug-302-redirect"},
-	{1, gen_cmd, num_rd, num_wr, 0, 1, &http_options.bug_post_no_keepalive, "http_bugs.bug_post_no_keepalive", "http-bugs.bug-post-no-keepalive"},
+	{1, gen_cmd, NULL, NULL, 0, 1, NULL, "http_bugs.bug_302_redirect", NULL},
+	{1, gen_cmd, NULL, NULL, 0, 1, NULL, "http_bugs.bug_post_no_keepalive", NULL},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &http_options.no_accept_charset, "http_bugs.no_accept_charset", "http-bugs.bug-no-accept-charset"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &http_options.no_compression, "http_bugs.no_compression", "http-bugs.no-compression"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &http_options.retry_internal_errors, "http_bugs.retry_internal_errors", "http-bugs.retry-internal-errors"},
@@ -2260,7 +2298,7 @@ static struct option links_options[] = {
 	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, ftp_options.anon_pass, "ftp.anonymous_password", "ftp.anonymous-password"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &ftp_options.passive_ftp, "ftp.use_passive", "ftp.use-passive"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &ftp_options.eprt_epsv, "ftp.use_eprt_epsv", "ftp.use-eprt-epsv"},
-	{1, gen_cmd, num_rd, num_wr, 0, 1, &ftp_options.fast_ftp, "ftp.fast", "ftp.fast"},
+	{1, gen_cmd, NULL, NULL, 0, 1, NULL, "ftp.fast", NULL},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &ftp_options.set_tos, "ftp.set_iptos", "ftp.set-iptos"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &smb_options.allow_hyperlinks_to_smb, "smb.allow_hyperlinks_to_smb", "smb.allow-hyperlinks-to-smb"},
 	{1, gen_cmd, num_rd, num_wr, 1, MAX_FONT_SIZE, &menu_font_size, "menu_font_size", "menu-font-size"},
@@ -2272,17 +2310,30 @@ static struct option links_options[] = {
 	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, bookmarks_file, "bookmarks_file", "bookmarks-file"},
 	{1, gen_cmd, cp_rd, cp_wr, 0, 0, &bookmarks_codepage, "bookmarks_codepage", "bookmarks-codepage"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &save_history, "save_url_history", "save-url-history"},
+	{1, gen_cmd, num_rd, num_wr, 0, 1, &enable_cookies, "enable_cookies", "enable-cookies"},
+	{1, gen_cmd, num_rd, num_wr, 0, 1, &save_cookies, "save_cookies", "save-cookies"},
+	{1, gen_cmd, dbl_rd, dbl_wr, 0, 999999900, &max_cookie_age, "max_cookie_age", "max-cookie-age"},
 	{1, gen_cmd, dbl_rd, dbl_wr, 1, 10000, &display_red_gamma, "display_red_gamma", "display-red-gamma"},
 	{1, gen_cmd, dbl_rd, dbl_wr, 1, 10000, &display_green_gamma, "display_green_gamma", "display-green-gamma"},
 	{1, gen_cmd, dbl_rd, dbl_wr, 1, 10000, &display_blue_gamma, "display_blue_gamma", "display-blue-gamma"},
 	{1, gen_cmd, dbl_rd, dbl_wr, 1, 10000, &user_gamma, "user_gamma", "user-gamma"},
 	{1, gen_cmd, dbl_rd, dbl_wr, 25, 400, &bfu_aspect, "bfu_aspect", "bfu-aspect"},
-	{1, gen_cmd, num_rd, NULL, 0, 1, &sink, "aspect_on", NULL},
+	{1, gen_cmd, NULL, NULL, 0, 1, NULL, "aspect_on", NULL},
 	{1, gen_cmd, num_rd, num_wr, 0, 2, &display_optimize, "display_optimize", "display-optimize"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &dither_letters, "dither_letters", "dither-letters"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &dither_images, "dither_images", "dither-images"},
 	{1, gen_cmd, num_rd, num_wr, 0, 2, &gamma_bits, "gamma_correction", "gamma-correction"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &overwrite_instead_of_scroll, "overwrite_instead_of_scroll", "overwrite-instead-of-scroll"},
+	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, font_file, "font", "font"},
+	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, font_file_b, "font_bold", "font-bold"},
+	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, font_file_m, "font_monospaced", "font-monospaced"},
+	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, font_file_m_b, "font_monospaced_bold", "font-monospaced-bold"},
+#ifdef USE_ITALIC
+	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, font_file_i, "font_italic", "font-italic"},
+	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, font_file_i_b, "font_italic_bold", "font-italic-bold"},
+	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, font_file_i_m, "font_monospaced_italic", "font-monospaced-italic"},
+	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, font_file_i_m_b, "font_monospaced_italic_bold", "font-monospaced-italic-bold"},
+#endif
 	{1, gen_cmd, num_rd, NULL, 0, 1, &js_enable, "enable_javascript", NULL},
 	{1, gen_cmd, num_rd, NULL, 0, 1, &js_verbose_errors, "verbose_javascript_errors", NULL},
 	{1, gen_cmd, num_rd, NULL, 0, 1, &js_verbose_warnings, "verbose_javascript_warnings", NULL},
@@ -2301,6 +2352,7 @@ static struct option links_options[] = {
 	{1, NULL, prog_rd, prog_wr, 0, 0, &tn3270_prog, "tn3270", NULL},
 	{1, NULL, prog_rd, prog_wr, 0, 0, &mms_prog, "mms", NULL},
 	{1, NULL, prog_rd, prog_wr, 0, 0, &magnet_prog, "magnet", NULL},
+	{1, NULL, prog_rd, prog_wr, 0, 0, &gopher_prog, "gopher", NULL},
 	{1, NULL, block_rd, block_wr, 0, 0, NULL, "imageblock", NULL},
 	{1, NULL, dp_rd, dp_wr, 0, 0, NULL, "video_driver", NULL},
 	{0, NULL, NULL, NULL, 0, 0, NULL, NULL, NULL},
@@ -2336,9 +2388,17 @@ static struct option html_options[] = {
 
 static struct option *all_options[] = { links_options, html_options, NULL, };
 
-unsigned char *parse_options(int argc, unsigned char *argv[])
+unsigned char *parse_options(int argc, char *argv[])
 {
-	return p_arse_options(argc, argv, all_options);
+	int i;
+	unsigned char **u_argv, *ret;
+	if ((unsigned)argc > MAXINT / sizeof(unsigned char *)) overalloc();
+	u_argv = mem_alloc(argc * sizeof(unsigned char *));
+	for (i = 0; i < argc; i++)
+		u_argv[i] = cast_uchar argv[i];
+	ret = p_arse_options(argc, u_argv, all_options);
+	mem_free(u_argv);
+	return ret;
 }
 
 static void load_config_file(unsigned char *prefix, unsigned char *name)
@@ -2436,6 +2496,5 @@ void save_url_history(void)
 	write_to_config_file(history_file, hs, 0);
 	mem_free(history_file);
 	mem_free(hs);
-	return;
 }
 
